#include <stdio.h>
#include <string.h>
#include "AVGA/avga.h"

//keyboard
#include "input.h"


//grahical files
#include "graphics/tileset.h"
#include "graphics/mario16.h"  //generated with x2h_rle
#include "graphics/map1.h"     //generated with x2h


//user graphical objects
const unsigned char anim1[] PROGMEM = MARIO16;
const unsigned char map1[] PROGMEM = MAP1;

//sound files
//#include "sounds/music1.h"
#include "sounds/sounds.h"


//user sound objects
const unsigned char snd_jump[] PROGMEM = SND_JUMP;
const unsigned char snd_tram[] PROGMEM = SND_TRAM;
const unsigned char snd_impact[] PROGMEM = SND_IMPACT;
const unsigned char snd_fall[] PROGMEM = SND_FALL;


//memory for the driver.
unsigned char screen[DRIVER_REFTABLE_SIZE];
const unsigned char pgmmap[] PROGMEM = TILESET;
unsigned char rammap[DRIVER_RAMMAP_SIZE];


//Some high-level helpful functions.
static unsigned char RenderMap( signed int x, signed int y );
static void DrawFrame( unsigned char x, unsigned char y, unsigned char width, unsigned char height, unsigned char frame );
static void DrawInt16( int x, int y, unsigned int i );




#define MARIO_WIDTH	16
#define MARIO_HEIGHT 16



int main( void )
{
	//map memory for video driver
	driver_mmap(screen, pgmmap, rammap);

	//initialization
	sound_init();
	video_init(NTSC);
	input_init();

	enable(); 			//start engine.
	wait_seconds(1); 	//wait for display to sync.

	sync();
	//sound_play(music1);

	driver_fill(128);
	wait_seconds(1);
	driver_print_P(6, 1, PSTR("JD presents:")); 
	driver_print_P(5, 3, PSTR("AVR SuperMario"));
	driver_print_P(5, 4, PSTR("=============="));
	driver_print_P(5, 6, PSTR("(AlphaVersion)"));
	driver_print_P(8, 9, PSTR("Based on"));
	driver_print_P(3, 11, PSTR("AVGA video driver."));
	driver_print_P(6, 13, PSTR("md@unicode.cz"));
	driver_print_P(1, 16, PSTR("PRESS ANY KEY TO START"));
	//wait_seconds(6);

	//overlay_draw_sprite(3,5,1);

	key_get_event();
	while(!key_get_event()) sync();




	driver_fill(128);
	WINDOW wnd1 = window_add_standard(0, 15);
	WINDOW wnd2 = window_add_standard(16, 3); //leave one row for scroll.

	window_select(wnd2);
	DrawFrame(0, 0, 22, 1, 24); //5
	


	unsigned char g=1, a=1, j=0,  jump=0, air=0, u=1;
//	signed char ldx, vx=0;//by
	signed char ldx=0, vx=0;

	signed char dx=0, dy=0;
	unsigned int x=100, y=100; 		//x a y jsou souradnice postavy vzhledem k levemu hornimu okraji mapy. VZDY
	signed int xm=0, ym=0;

	PGM_P frame = anim1; //animation

	while(1)
	{
		//render();
		window_select(wnd1); 

		if(RenderMap(xm, ym))
			overlay_release();
		else
			overlay_clear();

		unsigned int xb, yb; //poloha postavy na obrazovce.
		overlay_draw_param(xb = x+xm, yb = y+ym, frame, 16, 16, ldx < 0 ? OVERLAY_HFLIP : 0); //nebo: xs, ys
		

    	//frameMove();

		//user input
		unsigned char event = key_get_event();
		if(event & KEY_UP && !air) { sound_play(snd_jump); jump = 1; dy = -3; }
		if(event & KEY_LEFT)  a=1;
		if(event & KEY_RIGHT) a=1;


		unsigned char status = key_get_status();
		vx=0;
		if(status & KEY_RIGHT) vx+=2; //desired speed
		if(status & KEY_LEFT)  vx-=2; //desired speed
		if(status & KEY_DOWN)  ym--;
		if(vx) ldx=vx;

		if(!--a) 
		{ 
			//dx+=sgn(vx-dx);
			if(vx>dx)dx++;
			if(vx<dx)dx--; 
			a=5; 				//1/acceleration of Mario.
		}
		

		//collision engine.
		
		//note that absolute value of dx and dy must always be lower or equal to 8.
		//otherwise the collision detector would be much more complicated to work properly.
		//ziskat kolizni blok. bereme ze SOUCASNE VIDITELNE MAPY, dopocitavame absolutni polohu.
		//#define window_coords2block() ...
		//xb a yb jsou ted absolutni souradnice leveho horniho bloku MARIA NA OBRAZOVCE.


		xb += window_get_scrollX();
		yb += window_get_scrollY();  //window_get_startline();
		xb >>= 3;
		yb >>= 3;

		unsigned char colwidth = 2;
		unsigned char colheight = 2;
		if(x&7) colwidth++;
		if(y&7) colheight++;


vert:	if(dy > 0)  //dole
		{
				   y--;
				   unsigned char pre = y&0xF8;
				   y += dy;
				   unsigned char post = y&0xF8;
				   y++;
				   if(post == pre) goto horiz;

				   unsigned char k = yb + colheight;
				   unsigned char l = xb;
				   for(j=0;j<colwidth;j++) 					//projet radek.
				   {
				     unsigned char b=window_get_block(l++,k);
					 if(b > 212 && b < 217 || b == 222 || b == 150) 		//optimalizovat pomoci pointeru na block
					 { 
					  	dy=0;

						if(air) sound_play(snd_impact);
						air = 0; 		  //muzeme skocit?
						jump = 0;		  //pro animaci
						g = 1;			  //gravity timer
						if(b==222) { sound_play(snd_tram); dy=-6;} //trampolina

						//y &= 0xFFF8;
						asm volatile("mov %A0, %1" : "=r" (y) : "r" (post));
						goto horiz; 
					 }
				   }
				   //k vertikalni kolizi nedoslo - posunout blok pro horizontalni kolizi
				   //yb++;
				   colheight++;
		}
	    else if(dy < 0) //nahore
		{
				  unsigned char pre = y&0xF8;
				  y += dy;
				  unsigned char post = y&0xF8;
				  if(post == pre) goto horiz;

				  unsigned char k = yb - 1;
				  unsigned char l = xb;

				  for(j=0;j<colwidth;j++) 						//projet radek.
				  {
				    unsigned char b=window_get_block(l++,k);
					if(b > 212 && b < 217) 
					{ 
						dy = 0; 
						asm volatile("mov %A0, %1" : "=r" (y) : "r" (pre));
						goto horiz; 
					}
				  }

				  //k vertikalni kolizi nedoslo - posunout blok pro horizontalni kolizi
				  yb--;
				  colheight++;
		}
horiz: 	if(dx > 0) //vpravo
		{
				  x--;
				  unsigned char pre = x&0xF8;
				  x += dx;
				  unsigned char post = x&0xF8;
				  x++;
				  if(post == pre) goto next;

				  unsigned char k = xb + colwidth;
				  unsigned char l = yb;
				  for(j=0;j<colheight;j++) 					//projet sloupecek.
				  {
				 	unsigned char b=window_get_block(k,l++);
					if(b > 212 && b < 217) 
					{ 
						dx = 0; 
						//x &= 0xFFF8;
						asm volatile("mov %A0, %1" : "=r" (x) : "r" (post));
						goto next; 
					}
				  }
		}
		else if(dx < 0) //vlevo
		{
				  unsigned char pre = x&0xF8;
				  x += dx;
				  unsigned char post = x&0xF8;
				  if(post == pre) goto next;

				  unsigned char k = xb - 1;
				  unsigned char l = yb;
				  for(j=0;j<colheight;j++) 					//projet sloupecek.
				  {
				  	unsigned char b=window_get_block(k,l++);
					if(b > 212 && b < 217) 
					{
					 	dx = 0; 
						asm volatile("mov %A0, %1" : "=r" (x) : "r" (pre));
						goto next; 
					}
				  }
		}

next:	if(dy>0 && !air) sound_play(snd_fall);
		if(dy) air=1;
		


		//gravitace, je vyhodnejsi tady nez nahore.
		if(!--g) 
		{
			if(dy<8) dy++;
			g=5;			//1/acceleration of Earth
		}

		//animace postavy. TODO:!!!!!
		if(!--u)
		{
			u=4;
			frame += 128;
			if((unsigned int)frame >= (unsigned int)anim1+(128*5))  frame = anim1 + 128;
		}
		if(!dx) frame = anim1;
		if(abs(dx-vx) > 1) frame = anim1 + 128*5;
		if(air) frame = anim1 + 128*4;
		if(jump) frame = anim1 + 128*6;


		//if(dx && !air) { aaa++; if((aaa&7) == 0) sound_play(aaa&8 ? snd_walk1:snd_walk2);  }
		


		//pohyb mapy vzhledem k postave. exponenciala
		signed int tmp2 = (((DRIVER_RESX/2)-(MARIO_WIDTH/2)))   - x;
		xm += (tmp2 - xm)/8;
   	    tmp2 = (((DRIVER_RESY/2)-(MARIO_HEIGHT/2)))  - y; 
		ym += (tmp2 - ym)/8;
		if(tmp2 - ym < -32) ym += (tmp2-ym) + 32;



		//ok, wait for retrace.


		
		window_select(wnd2);
		window_print_P(1, 1, PSTR("x=          y=       "));
		DrawInt16(3, 1, x); //show remaining time
		DrawInt16(15, 1, y);

		sync();
	}/**/
}







static void DrawInt16(int x, int y, unsigned int i)
{
	unsigned char str[6];
	itoa(i,str,10);
	window_print(x, y, str);
}

static void DrawFrame(unsigned char x, unsigned char y, unsigned char width, unsigned char height, unsigned char frame)
{
	unsigned char i;

	window_set_block(x++, y, frame++);
	for(i=width; i;i--) window_set_block(x++, y, frame); frame++;
	window_set_block(x, y++, frame++);
	for(i=height;i;i--) window_set_block(x, y++, frame); frame++;
	window_set_block(x--, y, frame++);
	for(i=width; i;i--) window_set_block(x--, y, frame); frame++;
	window_set_block(x, y--, frame++);
	for(i=height;i;i--) window_set_block(x, y--, frame); frame++;
}





static unsigned char RenderMap(signed int x, signed int y)
{
    x+=7;
	y+=7;

	window_adjust_for_background(x, y); 	//adjust window
	x>>=3;
	y>>=3;
	
	static signed char lasttmpx, lasttmpy;
	if(x == lasttmpx && y == lasttmpy) return 0; //no need to redraw!
	lasttmpx=x;
	lasttmpy=y;


	//now draw game map and everything attached to it.
	window_fill_scroll(128);


	background_draw( x, y, (PGM_P)map1, 60, 30 );
	background_draw( x+60, y, (PGM_P)map1, 60, 30 );
	background_draw( x+120, y, (PGM_P)map1, 60, 30 );



	//background_draw( x+30, y+20, avrlogo, 4, 2);

	return 1;
}

